#include <msp/core/algorithm.h>
#include <msp/core/maputils.h>
#include <msp/io/print.h>
#include "category.h"
#include "filter.h"
#include "poefilter.h"

using namespace std;
using namespace Msp;

FilterStatement::FilterStatement():
	show(true)
{ }

void FilterStatement::set_show(bool s)
{
	show = s;
}

void FilterStatement::add_line(const string &l)
{
	lines.push_back(l);
}

void FilterStatement::add_line(list<FilterStatement> &st, const string &line)
{
	for(list<FilterStatement>::iterator i=st.begin(); i!=st.end(); ++i)
		i->add_line(line);
}

void FilterStatement::write(IO::Base &out) const
{
	if(show)
		out.write("Show\n");
	else
		out.write("Hide\n");

	for(list<string>::const_iterator i=lines.begin(); i!=lines.end(); ++i)
		IO::print(out, "\t%s\n", *i);
}


Filter::Filter():
	abstract(false)
{ }

void Filter::write(IO::Base &out) const
{
	for(list<Block>::const_iterator i=blocks.begin(); i!=blocks.end(); ++i)
	{
		list<FilterStatement> st;
		st.push_back(FilterStatement());
		st.back().set_show(i->show);
		i->appearance.add_lines(st.back());
		i->category->create_statements(st);

		IO::print(out, "# %s\n", i->category->get_name());
		for(list<FilterStatement>::const_iterator j=st.begin(); j!=st.end(); ++j)
			j->write(out);
	}
}


Filter::Loader::Loader(Filter &f, const PoeFilter &p):
	DataFile::ObjectLoader<Filter>(f),
	poe(p)
{
	add("abstract", &Filter::abstract);
	add("cancel", &Loader::cancel);
	add("hide", &Loader::hide);
	add("include", &Loader::include);
	add("show", &Loader::show);
}

string Filter::Loader::glob_to_re(const string &glob)
{
	string result = "^";
	for(string::const_iterator i=glob.begin(); i!=glob.end(); ++i)
	{
		if(*i=='*')
			result += '.';
		else if(!isalnum(*i))
			result += '\\';
		result += *i;
	}
	result += '$';

	return result;
}

bool Filter::Loader::category_order(const Category *c1, const Category *c2)
{
	return c1->get_order()<c2->get_order();
}

void Filter::Loader::add_categories(const string &name, bool show)
{
	vector<const Category *> categs;
	if(name.find('*')!=string::npos)
	{
		poe.find_categories(glob_to_re(name), categs);

		if(categs.empty())
			throw key_error(name);

		sort(categs, &category_order);
	}
	else
		categs.push_back(&poe.get_category(name));

	for(vector<const Category *>::const_iterator i=categs.begin(); i!=categs.end(); ++i)
	{
		bool found = false;
		for(list<Block>::const_iterator j=obj.blocks.begin(); (!found && j!=obj.blocks.end()); ++j)
			found = (j->category==*i);

		if(!found)
		{
			Block blk;
			blk.show = show;
			blk.category = *i;
			blk.appearance = (*i)->get_appearance();
			if(const Appearance *app = poe.get_theme().find_appearance((*i)->get_name()))
				blk.appearance.merge_from(*app);
			obj.blocks.push_back(blk);
		}
	}
}

void Filter::Loader::cancel(const string &name)
{
	vector<const Category *> categs;
	if(name.find('*')!=string::npos)
	{
		poe.find_categories(glob_to_re(name), categs);

		if(categs.empty())
			throw key_error(name);
	}
	else
		categs.push_back(&poe.get_category(name));

	for(vector<const Category *>::const_iterator i=categs.begin(); i!=categs.end(); ++i)
		for(list<Block>::iterator j=obj.blocks.begin(); j!=obj.blocks.end(); ++j)
			if(j->category==*i)
			{
				obj.blocks.erase(j);
				break;
			}
}

void Filter::Loader::hide(const string &name)
{
	add_categories(name, false);
}

void Filter::Loader::include(const string &name)
{
	const Filter &base = poe.get_filter(name);
	obj.blocks.insert(obj.blocks.end(), base.blocks.begin(), base.blocks.end());
}

void Filter::Loader::show(const string &name)
{
	add_categories(name, true);
}
