#ifndef RANGECONDITION_H_
#define RANGECONDITION_H_

#include "condition.h"
#include "filter.h"

template<typename T>
class RangeCondition: public Condition
{
public:
	typedef T Traits;
	typedef typename Traits::Type Type;

protected:
	Type min;
	Type max;

public:
	RangeCondition(Type);
	RangeCondition(Type, Type);

	virtual RangeCondition<T> *clone() const;
	virtual bool equals(const Condition &) const;
	virtual bool can_merge(const Condition &, const CompoundCondition &) const;
	virtual RangeCondition<T> *merge(const std::vector<const Condition *> &, const CompoundCondition &) const;
	virtual bool is_viable() const { return min<=max; }
	virtual void add_lines(std::list<FilterStatement> &) const;
};


struct RarityTraits
{
	typedef Rarity Type;
	static Rarity get_min() { return NORMAL; }
	static Rarity get_max() { return UNIQUE; }
	static const char *get_keyword() { return "Rarity"; }
};

typedef RangeCondition<RarityTraits> RarityCondition;


struct QualityTraits
{
	typedef unsigned Type;
	static unsigned get_min() { return 0; }
	static unsigned get_max() { return 20; }
	static const char *get_keyword() { return "Quality"; }
};

typedef RangeCondition<QualityTraits> QualityCondition;


struct ItemLevelTraits
{
	typedef unsigned Type;
	static unsigned get_min() { return 1; }
	static unsigned get_max() { return 100; }
	static const char *get_keyword() { return "ItemLevel"; }
};

typedef RangeCondition<ItemLevelTraits> ItemLevelCondition;


struct DropLevelTraits
{
	typedef unsigned Type;
	static unsigned get_min() { return 1; }
	static unsigned get_max() { return 100; }
	static const char *get_keyword() { return "DropLevel"; }
};

typedef RangeCondition<DropLevelTraits> DropLevelCondition;


struct WidthTraits
{
	typedef unsigned Type;
	static unsigned get_min() { return 1; }
	static unsigned get_max() { return 2; }
	static const char *get_keyword() { return "Width"; }
};

typedef RangeCondition<WidthTraits> WidthCondition;


struct HeightTraits
{
	typedef unsigned Type;
	static unsigned get_min() { return 1; }
	static unsigned get_max() { return 4; }
	static const char *get_keyword() { return "Height"; }
};

typedef RangeCondition<HeightTraits> HeightCondition;


struct SocketsTraits
{
	typedef unsigned Type;
	static unsigned get_min() { return 1; }
	static unsigned get_max() { return 6; }
	static const char *get_keyword() { return "Sockets"; }
};

typedef RangeCondition<SocketsTraits> SocketsCondition;


struct LinkedSocketsTraits
{
	typedef unsigned Type;
	static unsigned get_min() { return 1; }
	static unsigned get_max() { return 6; }
	static const char *get_keyword() { return "LinkedSockets"; }
};

typedef RangeCondition<LinkedSocketsTraits> LinkedSocketsCondition;


template<typename T>
RangeCondition<T>::RangeCondition(Type v):
	min(v),
	max(v)
{ }

template<typename T>
RangeCondition<T>::RangeCondition(Type n, Type x):
	min(n),
	max(x)
{ }

template<typename T>
RangeCondition<T> *RangeCondition<T>::clone() const
{
	return new RangeCondition<T>(min, max);
}

template<typename T>
bool RangeCondition<T>::equals(const Condition &other) const
{
	const RangeCondition<T> *other_range = dynamic_cast<const RangeCondition<T> *>(&other);
	if(!other_range)
		return false;

	return (min==other_range->min && max==other_range->max);
}

template<typename T>
bool RangeCondition<T>::can_merge(const Condition &other, const CompoundCondition &parent) const
{
	const RangeCondition<T> *other_range = dynamic_cast<const RangeCondition<T> *>(&other);
	if(!other_range)
		return false;

	if(dynamic_cast<const AndCondition *>(&parent))
		return true;
	else if(dynamic_cast<const OrCondition *>(&parent))
		return min<=other_range->max+1 && max+1>=other_range->min;
	else
		return false;
}

template<typename T>
RangeCondition<T> *RangeCondition<T>::merge(const std::vector<const Condition *> &conditions, const CompoundCondition &parent) const
{
	if(conditions.empty())
		return 0;

	bool intersect = dynamic_cast<const AndCondition *>(&parent);
	RangeCondition<T> *result;
	if(intersect)
		result = new RangeCondition<T>(Traits::get_min(), Traits::get_max());
	else
		result = new RangeCondition<T>(Traits::get_max(), Traits::get_min());
	for(std::vector<const Condition *>::const_iterator i=conditions.begin(); i!=conditions.end(); ++i)
	{
		const RangeCondition<T> *c = static_cast<const RangeCondition<T> *>(*i);
		if(intersect)
		{
			result->min = std::max(result->min, c->min);
			result->max = std::min(result->max, c->max);
		}
		else
		{
			result->min = std::min(result->min, c->min);
			result->max = std::max(result->max, c->max);
		}
	}
	return result;
}

template<typename T>
void RangeCondition<T>::add_lines(std::list<FilterStatement> &st) const
{
	const char *keyword = Traits::get_keyword();
	if(min==max)
		FilterStatement::add_line(st, Msp::format("%s %s", keyword, min));
	else
	{
		if(min!=Traits::get_min())
			FilterStatement::add_line(st, Msp::format("%s >= %s", keyword, min));
		if(max!=Traits::get_max())
			FilterStatement::add_line(st, Msp::format("%s <= %s", keyword, max));
	}
}

#endif
